"""
    JS URLs compatibility utilities
    ===============================

    This modules defines utilities allowing to support multiple versions of Python and Django.

"""

import django


if django.VERSION < (2, 0):  # pragma: no cover
    from django.urls.resolvers import RegexURLPattern, RegexURLResolver
    RegexPattern = type('RegexPattern')
    RoutePattern = type('RoutePattern')
    URLResolver = RegexURLResolver
    URLPattern = RegexURLPattern
else:
    from django.urls.resolvers import (  # noqa: F401
        RegexPattern, RoutePattern, URLPattern, URLResolver,
    )



"""
    JS URLs serializer
    ==================

    This module defiine helper functions allowing to convert the URLs that are associated with the
    names or namespaces defined in the settings to JSON. The resulting JSON will be used by a
    Javascript module to provide reverse-like functionality.

"""

import json
import re
from urllib.parse import urljoin

from django.urls import get_resolver

from .conf import settings
from .utils.compat import RegexPattern, RoutePattern, URLPattern, URLResolver
from .utils.text import replace


url_arg_re = re.compile(r'(\(.*?\))')
url_kwarg_re = re.compile(r'(\(\?P\<(.*?)\>.*?\))')
url_optional_char_re = re.compile(r'(?:\w|/)(?:\?|\*)')
url_optional_group_re = re.compile(r'\(\?\:.*\)(?:\?|\*)')
url_path_re = re.compile(r'<(.*?)>')


def get_urls_as_json(resolver=None):
    """ Returns the URLs associated with an URL resover as JSON.

    This function will traverse the tree of URLs of an URL resolver in order to return a JSON
    containing the correspondance between fully qualified URL names and the related paths. Only URLs
    that are configured to be serialized are included in the final JSON.

    """
    resolver = resolver or get_resolver()
    return json.dumps(dict(_parse_resolver(resolver)))


def _parse_resolver(resolver, current_namespace=None, url_prefix=None, include_all=False):
    ns = (
        '{}:{}'.format(current_namespace, resolver.namespace)
        if current_namespace else resolver.namespace
    )
    include_all = include_all or (ns in settings.URLS)

    urls = []
    for url_pattern in resolver.url_patterns:
        if isinstance(url_pattern, URLResolver):
            new_url_prefix = _prepare_url_part(url_pattern)
            new_url_prefix = (
                '/' + new_url_prefix if url_prefix is None else
                urljoin(url_prefix or '/', new_url_prefix)
            )
            urls = urls + _parse_resolver(
                url_pattern,
                current_namespace=ns,
                url_prefix=new_url_prefix,
                include_all=include_all,
            )
        elif isinstance(url_pattern, URLPattern) and url_pattern.name:
            url_name = '{}:{}'.format(ns, url_pattern.name) if ns else url_pattern.name
            if url_name in settings.URLS or include_all:
                full_url = _prepare_url_part(url_pattern)
                urls.append((url_name, urljoin(url_prefix or '/', full_url)))

    return urls


def _prepare_url_part(url_pattern):
    url = ''

    if hasattr(url_pattern, 'regex'):  # pragma: no cover, NOTE: Django < 2.0 compatibility
        url = url_pattern.regex.pattern
    elif isinstance(url_pattern.pattern, RegexPattern):
        url = url_pattern.pattern._regex
    elif isinstance(url_pattern.pattern, RoutePattern):
        url = url_pattern.pattern._route

    final_url = replace(url, [('^', ''), ('$', '')])

    # Removes optional groups from the URL pattern.
    optional_group_matches = url_optional_group_re.findall(final_url)
    final_url = (
        replace(final_url, [(el, '') for el in optional_group_matches])
        if optional_group_matches else final_url
    )

    # Remvoves optional characters from the URL pattern.
    optional_char_matches = url_optional_char_re.findall(final_url)
    final_url = (
        replace(final_url, [(el, '') for el in optional_char_matches])
        if optional_char_matches else final_url
    )

    # Identifies named URL arguments inside the URL pattern.
    kwarg_matches = url_kwarg_re.findall(final_url)
    final_url = (
        replace(final_url, [(el[0], '<{}>'.format(el[1])) for el in kwarg_matches])
        if kwarg_matches else final_url
    )

    # Identifies unnamed URL arguments inside the URL pattern.
    args_matches = url_arg_re.findall(final_url)
    final_url = (
        replace(final_url, [(el, '<>') for el in args_matches]) if args_matches else final_url
    )

    # Identifies path expression and associated converters inside the URL pattern.
    path_matches = url_path_re.findall(final_url)
    final_url = (
        replace(final_url, [(el, el.split(':')[-1]) for el in path_matches])
        if (path_matches and not (kwarg_matches or args_matches)) else final_url
    )

    return final_url



"""
    JS URLs text utilities
    ======================

    This modules defines utilities allowing to text and strings.

"""

from functools import reduce


def replace(data, replacements):
    """ Allows to performs several string substitutions.

    This function performs several string substitutions on the initial ``data`` string using a list
    of 2-tuples (old, new) defining substitutions and returns the resulting string.

    """
    return reduce(lambda a, kv: a.replace(*kv), replacements, data)



"""
    JS URLs views
    =============

    This module defines views allowing to expose the Javascript helper and catalog of URLs allowing
    to provide reverse-like functionality on the client side.

"""

from django.views.generic import TemplateView

from .serializer import get_urls_as_json


class JsUrlsView(TemplateView):
    """ Renders a Javascript helper allowing to reverse Django URLs. """

    content_type = 'application/javascript'
    template_name = 'js_urls/js_urls.js'

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['urls'] = get_urls_as_json()
        return context



__version__ = '0.1.1'

# Deploying a new version:
# 1. remove the ".dev" from the current version number
# 2. create a new commit (eg. "Prepared 0.1.1 release")
# 3. run "git tag x.y.z" (eg. "git tag 0.1.1")
# 4. run "python setup.py sdist bdist_wheel upload"
# 5. bump the version (increment the version and append a ".dev" to it). eg. "0.1.2.dev"
# 6. create a new commit (eg. "Bumped version to 0.1.2.dev")
# 7. run "git push" and "git push --tags"


default_app_config = 'js_urls.apps.JSUrlsAppConfig'



"""
    JS URLs settings
    ================

    This file defines settings that can be overriden in the Django project's settings module.

"""

from django.conf import settings


# The "JS_URLS" setting allows to define which URLs should be serialized and made available in the
# Javascript helper. It should be noted that this setting should contain only URL names or
# namespaces. URL paths associated with configured URL names will be serialized. If namespaces are
# used in the context of this setting, all the underlying URL paths will be serialized.
URLS = getattr(settings, 'JS_URLS', [])



# -*- coding: utf-8 -*-

import codecs
from os.path import abspath
from os.path import dirname
from os.path import join
from setuptools import find_packages
from setuptools import setup

import js_urls


def read_relative_file(filename):
    """ Returns contents of the given file, whose path is supposed relative to this module. """
    with codecs.open(join(dirname(abspath(__file__)), filename), encoding='utf-8') as f:
        return f.read()


setup(
    name='django-js-urls',
    version=js_urls.__version__,
    author='impak Finance',
    author_email='tech@impakfinance.com',
    packages=find_packages(exclude=['tests.*', 'tests']),
    include_package_data=True,
    url='https://github.com/impak-finance/django-js-urls',
    license='MIT',
    description='A convenient helper to use Django URLs in Javascript.',
    long_description=read_relative_file('README.rst'),
    keywords='django javascript js url urls helper',
    zip_safe=False,
    install_requires=[
        'django>=1.11',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
